import os
import time
import ase.db
import numpy as np
from copy import deepcopy
from random import random
from ase.io import write
from ase.optimize import BFGS
from ase.calculators.emt import EMT
from ase.ga.data import DataConnection
from ase.ga.population import Population
from ase.ga.standard_comparators import InteratomicDistanceComparator
from ase.ga.cutandsplicepairing import CutAndSplicePairing
from ase.ga.utilities import closest_distances_generator
from ase.ga.utilities import get_all_atom_types
from ase.ga.offspring_creator import OperationSelector
from ase.ga.standardmutations import MirrorMutation
from ase.ga.standardmutations import RattleMutation
from ase.ga.standardmutations import PermutationMutation
from ase.ga.convergence import GenerationRepetitionConvergence
from compare_rms import comparingtoreference
from xtb import GFN1


# Change the following parameters to suit your needs
# Specify the maximum number of generations this script will run
# The number of candidates to be tested is Pop x Gen
population_size = 20
mutation_probability = 0.3
num_gens = 120 

# Import databases
gadata = 'gadb.db'
refdata = 'ref.db'

da = DataConnection(gadata)
dc = ase.db.connect(gadata)
daref = DataConnection(refdata)
ref = ase.db.connect(refdata)

# Initialize the different components of the GA
atom_numbers_to_optimize = da.get_atom_numbers_to_optimize()
n_to_optimize = len(atom_numbers_to_optimize)
slab = da.get_slab()
all_atom_types = get_all_atom_types(slab, atom_numbers_to_optimize)
blmin = closest_distances_generator(all_atom_types,
                                    ratio_of_covalent_radii=0.7)

comp = InteratomicDistanceComparator(n_top=n_to_optimize,
                                     pair_cor_cum_diff=0.015,
                                     pair_cor_max=0.7,
                                     dE=0.02,
                                     mic=False)

pairing = CutAndSplicePairing(slab, n_to_optimize, blmin)
mutations = OperationSelector([1., 1., 0.],
                              [MirrorMutation(blmin, n_to_optimize),
                               RattleMutation(blmin, n_to_optimize),
                               PermutationMutation(n_to_optimize)])

# create the population
population = Population(data_connection=da,
                        population_size=population_size,
                        comparator=comp)

# We form generations in this algorithm run and can therefore set
# a convergence criteria based on generations
cc = GenerationRepetitionConvergence(population, 5)

# We can also compare candidates to a reference database to see 
# if expensive calculations are required
compare_ref = comparingtoreference(slab,n_to_optimize)

# Set calculation parameters
unpaired_electrons = 0
electronic_temp = 0
calc = GFN1(electronic_temperature=electronic_temp)
#calc = EMT()


# Make initial magnetic moments
slab_mm = np.zeros(len(slab.get_atomic_numbers())) 
cluster_mm = np.ones(n_to_optimize) * unpaired_electrons / n_to_optimize
initial_mm = np.append(slab_mm , cluster_mm)

# Relax all unrelaxed structures (e.g. the starting population)
while da.get_number_of_unrelaxed_candidates() > 0:
    a = da.get_an_unrelaxed_candidate()
    hold = deepcopy(a)
    print('Relaxing starting candidate {0}'.format(a.info['confid']-1), flush=True)
    a.set_calculator(calc)
    a.set_initial_magnetic_moments(initial_mm)
    dyn = BFGS(a, trajectory=None, logfile='bfgs.out')
    start = time.time()
    dyn.run(fmax=0.05, steps = 450)
    end = time.time() - start
    a.info['key_value_pairs']['raw_score'] = -a.get_potential_energy()
    a.info['key_value_pairs']['time'] = end
    da.add_relaxed_step(a)
    daref.add_unrelaxed_candidate(hold, 'copy:yes')
    daref.add_relaxed_step(a)
population.update()

# Run untill num_gens is reached 
while da.get_generation_number() < num_gens:

    print('Generation {0}'.format(da.get_generation_number()), flush=True)
    if cc.converged():
        print('converged', flush=True)
        break
    
    
    candidate = len(list(dc.select(relaxed=1,generation=da.get_generation_number())))
    while candidate < population_size:
        candidate += 1
        print('Relaxing candidate {0}'.format(candidate), flush=True)
        
        # Pairing operation
        a1, a2 = population.get_two_candidates()
        a3, desc = pairing.get_new_individual([a1, a2])
        if a3 is None:
            print('Generation {0} Candidate {1} is None'.format(da.get_generation_number(),candidate), flush=True)
            candidate -= 1
            continue
        da.add_unrelaxed_candidate(a3, description=desc)

        # Check if we want to do a mutation
        if random() < mutation_probability:
            while True:
                a3_mut, desc = mutations.get_new_individual([a3])
                if a3_mut is not None:
                    da.add_unrelaxed_step(a3_mut, desc)
                    a3 = a3_mut
                    break
        
        # Check if we need to relax
        a4, used = compare_ref.compare(a3, refdata)
        
        if used == 0:
            # Relax the new candidate
            hold = deepcopy(a3)
            a3.set_calculator(calc)
            a3.set_initial_magnetic_moments(initial_mm)
            dyn = BFGS(a3, trajectory=None, logfile='bfgs.out')
            start = time.time()
            dyn.run(fmax=0.05,steps = 250 )
            end = time.time() - start
            try:                
                a3.info['key_value_pairs']['raw_score'] = -a3.get_potential_energy()
                a3.info['key_value_pairs']['time'] = end
                da.add_relaxed_step(a3)
                daref.add_unrelaxed_candidate(hold, 'mutation:')
                daref.add_relaxed_step(a3)
                                #population.update()
            except (RuntimeError, TypeError, NameError):
                    pass
                #calc.clean()
                
        if used == 1:
            # Add the reference candidate as the relaxed structure, switching
            # atoms positions for reference positions and relaxing
            # Configurations is not added to reference database 
            a3.set_calculator(calc)
            a3.set_positions(a4.get_positions())
            a3.set_initial_magnetic_moments(initial_mm)
            dyn = BFGS(a3, trajectory=None, logfile=None)
            start = time.time()
            dyn.run(fmax=0.05, steps=2)
            end = time.time() - start
            try:
                print('Adding copy from ref.db', flush=True)                
                a3.info['key_value_pairs']['raw_score'] = -a3.get_potential_energy()
                a3.info['key_value_pairs']['time'] = end
                a3.info['key_value_pairs']['origin'] = 'CopiedFromRef.db'
                da.add_relaxed_step(a3)
            except (RuntimeError, TypeError, NameError):
                    pass
                #calc.clean()    
            
    population.update()

#write('all_candidates.traj', da.get_all_relaxed_candidates())
all_rex=da.get_all_relaxed_candidates()
write('all_candidates.db', all_rex)
cur_pop = population.get_current_population()
write('current_pop.db', cur_pop)

